rm(list=ls())

# set the seed as the last three digits of your student ID
seed = 111

if (seed == 111) {
  stop("\nThe random generator seed is still set to its default value.\nEdit the script and change it with the last three digits of your student ID\n\n", 
       call. = FALSE)
}
set.seed(seed)

# number of points in the dataset
trainSetSize = 230

################################# MAIN CODE ###########################################
cat("
    
-------------------------------------------------------------------------------------
Welcome to the second ML demo/homework (2017/2018): linear regression

For this demo we will use the \"Los Angeles ozone\" dataset. For more information
about it, check the following URLs:
(info) https://web.stanford.edu/~hastie/ElemStatLearn/datasets/LAozone.info.txt
(data) https://web.stanford.edu/~hastie/ElemStatLearn/datasets/LAozone.data

Let us now load the dataset and inspect its contents...

")
invisible(readline(prompt = "Press [enter] to continue"))  
# uncomment the line below if you want to read the dataset from the Web
#LAozone = read.table("http://statweb.stanford.edu/~tibs/ElemStatLearn/datasets/LAozone.data",sep=",",head=T)
LAozone = read.table("LAozone.data",sep=",",head=T)
attach(LAozone)
pairs(LAozone)

cat("

The dataset contains values for the following variables in different days 
of the year:
")
cat(names(LAozone))
cat("

We will now try to fit different simple univariate linear regression models
trying to estimate the \"ozone\" response variable from the data.

")
invisible(readline(prompt = "Press [enter] to continue"))  

numFeats = length(names(LAozone))
for (i in 2:numFeats){
  cat("Now showing ozone ~ ", names(LAozone)[i],": ")
  fit = lm(ozone ~ LAozone[,i])
  cat("RSS=",sum((predict(fit)-ozone)^2),"\n")
  plot(LAozone[,i],ozone,xlab=names(LAozone)[i])
  abline(fit)
  invisible(readline(prompt = "Press [enter] to continue"))    
}

cat("
    
Check the RSS values for the different features: \"temp\" has the lowest one,
thus it is the best choice if we want to fit our model using only one variable.
Better results, though, can be obtained if we use more than one variable:
let us then keep \"temp\" fixed and try to add each of the other features
to a multivariate linear regression model.
    
")
invisible(readline(prompt = "Press [enter] to continue"))    

# take all the columns except the one with temp
for (i in c(2:4,6:10)){ 
  cat("ozone ~ temp + ", names(LAozone)[i],": ")
  fit = lm(ozone ~ temp + LAozone[,i])
  cat("RSS=",sum((predict(fit)-ozone)^2),"\n")  
}

cat("
    
Now, temp+ibh looks like the best choice. Let us choose a third variable...

")
invisible(readline(prompt = "Press [enter] to continue"))    

# take all the columns except the one with temp and the one with ibh
for (i in c(2:4,7:10)){ 
  cat("ozone ~ temp + ibh + ", names(LAozone)[i],": ")
  fit = lm(ozone ~ temp + ibh + LAozone[,i])
  cat("RSS=",sum((predict(fit)-ozone)^2),"\n")  
}

cat("
For three variables, temp+ibh+humidity is the best choice.

Now let us extend this approach to all the variables: the *regsubsets*
command from the *leaps* library allows you to do that automatically.
Check out its output below here:
    
")
invisible(readline(prompt = "Press [enter] to continue"))    
library(leaps)
regs = regsubsets(ozone ~., LAozone, method="backward")
summary(regs)


cat("

Q1) The variable selection results we obtain with the regsubsets command 
are consistent with the first three steps we performed manually. Can you 
tell which approach we used for variable selection (both in regsubsets 
and in manual selection)? Would the sets of selected variables be the same 
if we chose another approach in regsubsets?
(NOTE: you can try that! Just type ?regsubsets in R to see which methods 
are allowed, and change the current one in the code)

Q2) Which is the best variable selection method in qualitative terms? 
Choose e.g. the sets of 4 variables you get with the different methods 
provided by regsubsets, fit linear models using them, and show which 
one has the lowest MSE. Can you also explain why that method is better?

")

cat("

Let us now try to fit models up to 4 variables (rows 1-4 in the table 
above), plus another model containing all of them.

")
invisible(readline(prompt = "Press [enter] to continue"))    

fit1 = lm(ozone ~ temp)
fit2 = lm(ozone ~ temp + ibh)
fit3 = lm(ozone ~ temp + ibh + humidity)
fit4 = lm(ozone ~ temp + humidity + ibt + doy)
fitx = lm(ozone ~ ., data=LAozone)

cat("ozone ~ temp: MSE=", mean((predict(fit1)-ozone)^2),"\n")
cat("ozone ~ temp + ibh: MSE=", mean((predict(fit2)-ozone)^2),"\n")
cat("ozone ~ temp + ibh + humidity: MSE=", mean((predict(fit3)-ozone)^2),"\n")
cat("ozone ~ temp + humidity + ibt + doy: MSE=", mean((predict(fit4)-ozone)^2),"\n")
cat("ozone ~ ALL: MSE=", mean((predict(fitx)-ozone)^2),"\n")

cat("

As you can see, there is not much gain by choosing to use all the variables
with respect to choosing only four of them. Moreover, the model would be 
much more complicated than needed.
Rather, you might want to take into account some non-linearity. For instance,
check the result below:

")

fit5 = lm(ozone ~ temp + humidity + ibt + poly(doy,2))
cat("ozone ~ temp + humidity + ibt + poly(doy,2): MSE=", mean((predict(fit5)-ozone)^2),"\n")

cat("
    
This looks pretty cool: just by introducing a small nonlinear factor
we can get a better result than by taking all of the variables!

Now, the question we should ask ourselves now is: are we already 
overfitting? To verify this, let us split the dataset in a training 
and a test subset, and fit again our functions...

")
invisible(readline(prompt = "Press [enter] to continue"))    
    
dataItems = nrow(LAozone)
cat("\n\nTotal number of items in the dataset: ",dataItems,"\n")
cat("Training set size: ", trainSetSize, "\n")
cat("Test set size: ", dataItems-trainSetSize, "\n\n")
train = rep(FALSE,dataItems)
train[sample(1:330, trainSetSize)] = TRUE

LA.train = LAozone[train,]
LA.test = LAozone[!train,]

fit1 = lm(ozone ~ temp, data=LA.train)
fit2 = lm(ozone ~ temp + ibh, data = LA.train)
fit3 = lm(ozone ~ temp + ibh + humidity, data = LA.train)
fit4 = lm(ozone ~ temp + humidity + ibt + doy, data = LA.train)
fitx = lm(ozone ~ ., data = LA.train)
fit5 = lm(ozone ~ temp + humidity + ibt + poly(doy,2), data = LA.train)

cat("ozone ~ temp                               : MSE.train=", mean((predict(fit1,LA.train)-LA.train$ozone)^2), 
    " MSE.test=", mean((predict(fit1,LA.test)-LA.test$ozone)^2), "\n")
cat("ozone ~ temp + ibh                         : MSE.train=", mean((predict(fit2,LA.train)-LA.train$ozone)^2), 
    " MSE.test=", mean((predict(fit2,LA.test)-LA.test$ozone)^2), "\n")
cat("ozone ~ temp + ibh + humidity              : MSE.train=", mean((predict(fit3,LA.train)-LA.train$ozone)^2), 
    " MSE.test=", mean((predict(fit3,LA.test)-LA.test$ozone)^2), "\n")
cat("ozone ~ temp + humidity + ibt + doy        : MSE.train=", mean((predict(fit4,LA.train)-LA.train$ozone)^2), 
    " MSE.test=", mean((predict(fit4,LA.test)-LA.test$ozone)^2), "\n")
cat("ozone ~ ALL                                : MSE.train=", mean((predict(fitx,LA.train)-LA.train$ozone)^2), 
    " MSE.test=", mean((predict(fitx,LA.test)-LA.test$ozone)^2), "\n")
cat("ozone ~ temp + humidity + ibt + poly(doy,2): MSE.train=", mean((predict(fit5,LA.train)-LA.train$ozone)^2), 
    " MSE.test=", mean((predict(fit5,LA.test)-LA.test$ozone)^2), "\n")

cat("
To give you an idea about how well our function fits well the data, we now
plot the ozone variable as a function of temp. Black points belong to the
training set, green ones to test, and red ones are our predictions of the
test set.

You can play with the variables to see, for instance:
- how different training/test set sizes affect results
- how the plot changes if you use different functions (change the \"fit\" var)
- how the random initialization affects the training/test sets,
  and consequently the final results.
    
")


plot(LA.train$temp,LA.train$ozone)
points(LA.test$temp,LA.test$ozone,col="green")
points(LA.test$temp,predict(fit5,LA.test),col="red")

cat("

Q3) What can you say about the final results, comparing the evaluations 
of the training and test sets? Do you think there is actually overfitting? 
Do things change if you choose different values for the training and 
test sets?

Q4) Can you do better? Try to modify the code to include other types 
of non-linearities (in the simplest case, more complex polynomials) 
and verify whether the model gets better only in the training set 
or also in test.

")

detach(LAozone)

